# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2005,2007,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Official Cedar Backup Extensions
# Purpose  : Provides an extension to back up Subversion repositories.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Provides an extension to back up Subversion repositories.

This is a Cedar Backup extension used to back up Subversion repositories via
the Cedar Backup command line.  Each Subversion repository can be backed using
the same collect modes allowed for filesystems in the standard Cedar Backup
collect action: weekly, daily, incremental.

This extension requires a new configuration section <subversion> and is
intended to be run either immediately before or immediately after the standard
collect action.  Aside from its own configuration, it requires the options and
collect configuration sections in the standard Cedar Backup configuration file.

There are two different kinds of Subversion repositories at this writing: BDB
(Berkeley Database) and FSFS (a "filesystem within a filesystem").  Although
the repository type can be specified in configuration, that information is just
kept around for reference.  It doesn't affect the backup.  Both kinds of
repositories are backed up in the same way, using ``svnadmin dump`` in an
incremental mode.

It turns out that FSFS repositories can also be backed up just like any
other filesystem directory.  If you would rather do that, then use the normal
collect action.  This is probably simpler, although it carries its own
advantages and disadvantages (plus you will have to be careful to exclude
the working directories Subversion uses when building an update to commit).
Check the Subversion documentation for more information.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules
########################################################################

import logging
import os
import pickle
import posixpath
from bz2 import BZ2File
from functools import total_ordering
from gzip import GzipFile

from CedarBackup3.config import VALID_COLLECT_MODES, VALID_COMPRESS_MODES
from CedarBackup3.filesystem import FilesystemList
from CedarBackup3.util import (
    ObjectTypeList,
    RegexList,
    UnorderedList,
    buildNormalizedPath,
    changeOwnership,
    encodePath,
    executeCommand,
    isStartOfWeek,
    pathJoin,
    resolveCommand,
)
from CedarBackup3.xmlutil import (
    addContainerNode,
    addStringNode,
    createInputDom,
    isElement,
    readChildren,
    readFirstChild,
    readString,
    readStringList,
)

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.extend.subversion")

SVNLOOK_COMMAND = ["svnlook"]
SVNADMIN_COMMAND = ["svnadmin"]

REVISION_PATH_EXTENSION = "svnlast"


########################################################################
# RepositoryDir class definition
########################################################################


@total_ordering
class RepositoryDir(object):
    """
    Class representing Subversion repository directory.

    A repository directory is a directory that contains one or more Subversion
    repositories.

    The following restrictions exist on data in this class:

       - The directory path must be absolute.
       - The collect mode must be one of the values in :any:`VALID_COLLECT_MODES`.
       - The compress mode must be one of the values in :any:`VALID_COMPRESS_MODES`.

    The repository type value is kept around just for reference.  It doesn't
    affect the behavior of the backup.

    Relative exclusions are allowed here.  However, there is no configured
    ignore file, because repository dir backups are not recursive.

    """

    def __init__(
        self,
        repositoryType=None,
        directoryPath=None,
        collectMode=None,
        compressMode=None,
        relativeExcludePaths=None,
        excludePatterns=None,
    ):
        """
        Constructor for the ``RepositoryDir`` class.

        Args:
           repositoryType: Type of repository, for reference
           directoryPath: Absolute path of the Subversion parent directory
           collectMode: Overridden collect mode for this directory
           compressMode: Overridden compression mode for this directory
           relativeExcludePaths: List of relative paths to exclude
           excludePatterns: List of regular expression patterns to exclude
        """
        self._repositoryType = None
        self._directoryPath = None
        self._collectMode = None
        self._compressMode = None
        self._relativeExcludePaths = None
        self._excludePatterns = None
        self.repositoryType = repositoryType
        self.directoryPath = directoryPath
        self.collectMode = collectMode
        self.compressMode = compressMode
        self.relativeExcludePaths = relativeExcludePaths
        self.excludePatterns = excludePatterns

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "RepositoryDir(%s, %s, %s, %s, %s, %s)" % (
            self.repositoryType,
            self.directoryPath,
            self.collectMode,
            self.compressMode,
            self.relativeExcludePaths,
            self.excludePatterns,
        )

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.repositoryType != other.repositoryType:
            if str(self.repositoryType or "") < str(other.repositoryType or ""):
                return -1
            else:
                return 1
        if self.directoryPath != other.directoryPath:
            if str(self.directoryPath or "") < str(other.directoryPath or ""):
                return -1
            else:
                return 1
        if self.collectMode != other.collectMode:
            if str(self.collectMode or "") < str(other.collectMode or ""):
                return -1
            else:
                return 1
        if self.compressMode != other.compressMode:
            if str(self.compressMode or "") < str(other.compressMode or ""):
                return -1
            else:
                return 1
        if self.relativeExcludePaths != other.relativeExcludePaths:
            if self.relativeExcludePaths < other.relativeExcludePaths:
                return -1
            else:
                return 1
        if self.excludePatterns != other.excludePatterns:
            if self.excludePatterns < other.excludePatterns:
                return -1
            else:
                return 1
        return 0

    def _setRepositoryType(self, value):
        """
        Property target used to set the repository type.
        There is no validation; this value is kept around just for reference.
        """
        self._repositoryType = value

    def _getRepositoryType(self):
        """
        Property target used to get the repository type.
        """
        return self._repositoryType

    def _setDirectoryPath(self, value):
        """
        Property target used to set the directory path.
        The value must be an absolute path if it is not ``None``.
        It does not have to exist on disk at the time of assignment.
        Raises:
           ValueError: If the value is not an absolute path
           ValueError: If the value cannot be encoded properly
        """
        if value is not None:
            if not (os.path.isabs(value) or posixpath.isabs(value)):  # Python 3.13+ does not treat / as absolute on Windows
                raise ValueError("Repository path must be an absolute path.")
        self._directoryPath = encodePath(value)

    def _getDirectoryPath(self):
        """
        Property target used to get the repository path.
        """
        return self._directoryPath

    def _setCollectMode(self, value):
        """
        Property target used to set the collect mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_COLLECT_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_COLLECT_MODES:
                raise ValueError("Collect mode must be one of %s." % VALID_COLLECT_MODES)
        self._collectMode = value

    def _getCollectMode(self):
        """
        Property target used to get the collect mode.
        """
        return self._collectMode

    def _setCompressMode(self, value):
        """
        Property target used to set the compress mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_COMPRESS_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_COMPRESS_MODES:
                raise ValueError("Compress mode must be one of %s." % VALID_COMPRESS_MODES)
        self._compressMode = value

    def _getCompressMode(self):
        """
        Property target used to get the compress mode.
        """
        return self._compressMode

    def _setRelativeExcludePaths(self, value):
        """
        Property target used to set the relative exclude paths list.
        Elements do not have to exist on disk at the time of assignment.
        """
        if value is None:
            self._relativeExcludePaths = None
        else:
            try:
                saved = self._relativeExcludePaths
                self._relativeExcludePaths = UnorderedList()
                self._relativeExcludePaths.extend(value)
            except Exception as e:
                self._relativeExcludePaths = saved
                raise e

    def _getRelativeExcludePaths(self):
        """
        Property target used to get the relative exclude paths list.
        """
        return self._relativeExcludePaths

    def _setExcludePatterns(self, value):
        """
        Property target used to set the exclude patterns list.
        """
        if value is None:
            self._excludePatterns = None
        else:
            try:
                saved = self._excludePatterns
                self._excludePatterns = RegexList()
                self._excludePatterns.extend(value)
            except Exception as e:
                self._excludePatterns = saved
                raise e

    def _getExcludePatterns(self):
        """
        Property target used to get the exclude patterns list.
        """
        return self._excludePatterns

    repositoryType = property(_getRepositoryType, _setRepositoryType, None, doc="Type of this repository, for reference.")
    directoryPath = property(_getDirectoryPath, _setDirectoryPath, None, doc="Absolute path of the Subversion parent directory.")
    collectMode = property(_getCollectMode, _setCollectMode, None, doc="Overridden collect mode for this repository.")
    compressMode = property(_getCompressMode, _setCompressMode, None, doc="Overridden compress mode for this repository.")
    relativeExcludePaths = property(_getRelativeExcludePaths, _setRelativeExcludePaths, None, "List of relative paths to exclude.")
    excludePatterns = property(_getExcludePatterns, _setExcludePatterns, None, "List of regular expression patterns to exclude.")


########################################################################
# Repository class definition
########################################################################


@total_ordering
class Repository(object):
    """
    Class representing generic Subversion repository configuration..

    The following restrictions exist on data in this class:

       - The respository path must be absolute.
       - The collect mode must be one of the values in :any:`VALID_COLLECT_MODES`.
       - The compress mode must be one of the values in :any:`VALID_COMPRESS_MODES`.

    The repository type value is kept around just for reference.  It doesn't
    affect the behavior of the backup.

    """

    def __init__(self, repositoryType=None, repositoryPath=None, collectMode=None, compressMode=None):
        """
        Constructor for the ``Repository`` class.

        Args:
           repositoryType: Type of repository, for reference
           repositoryPath: Absolute path to a Subversion repository on disk
           collectMode: Overridden collect mode for this directory
           compressMode: Overridden compression mode for this directory
        """
        self._repositoryType = None
        self._repositoryPath = None
        self._collectMode = None
        self._compressMode = None
        self.repositoryType = repositoryType
        self.repositoryPath = repositoryPath
        self.collectMode = collectMode
        self.compressMode = compressMode

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "Repository(%s, %s, %s, %s)" % (self.repositoryType, self.repositoryPath, self.collectMode, self.compressMode)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.repositoryType != other.repositoryType:
            if str(self.repositoryType or "") < str(other.repositoryType or ""):
                return -1
            else:
                return 1
        if self.repositoryPath != other.repositoryPath:
            if str(self.repositoryPath or "") < str(other.repositoryPath or ""):
                return -1
            else:
                return 1
        if self.collectMode != other.collectMode:
            if str(self.collectMode or "") < str(other.collectMode or ""):
                return -1
            else:
                return 1
        if self.compressMode != other.compressMode:
            if str(self.compressMode or "") < str(other.compressMode or ""):
                return -1
            else:
                return 1
        return 0

    def _setRepositoryType(self, value):
        """
        Property target used to set the repository type.
        There is no validation; this value is kept around just for reference.
        """
        self._repositoryType = value

    def _getRepositoryType(self):
        """
        Property target used to get the repository type.
        """
        return self._repositoryType

    def _setRepositoryPath(self, value):
        """
        Property target used to set the repository path.
        The value must be an absolute path if it is not ``None``.
        It does not have to exist on disk at the time of assignment.
        Raises:
           ValueError: If the value is not an absolute path
           ValueError: If the value cannot be encoded properly
        """
        if value is not None:
            if not (os.path.isabs(value) or posixpath.isabs(value)):  # Python 3.13+ does not treat / as absolute on Windows
                raise ValueError("Repository path must be an absolute path.")
        self._repositoryPath = encodePath(value)

    def _getRepositoryPath(self):
        """
        Property target used to get the repository path.
        """
        return self._repositoryPath

    def _setCollectMode(self, value):
        """
        Property target used to set the collect mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_COLLECT_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_COLLECT_MODES:
                raise ValueError("Collect mode must be one of %s." % VALID_COLLECT_MODES)
        self._collectMode = value

    def _getCollectMode(self):
        """
        Property target used to get the collect mode.
        """
        return self._collectMode

    def _setCompressMode(self, value):
        """
        Property target used to set the compress mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_COMPRESS_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_COMPRESS_MODES:
                raise ValueError("Compress mode must be one of %s." % VALID_COMPRESS_MODES)
        self._compressMode = value

    def _getCompressMode(self):
        """
        Property target used to get the compress mode.
        """
        return self._compressMode

    repositoryType = property(_getRepositoryType, _setRepositoryType, None, doc="Type of this repository, for reference.")
    repositoryPath = property(_getRepositoryPath, _setRepositoryPath, None, doc="Path to the repository to collect.")
    collectMode = property(_getCollectMode, _setCollectMode, None, doc="Overridden collect mode for this repository.")
    compressMode = property(_getCompressMode, _setCompressMode, None, doc="Overridden compress mode for this repository.")


########################################################################
# SubversionConfig class definition
########################################################################


@total_ordering
class SubversionConfig(object):
    """
    Class representing Subversion configuration.

    Subversion configuration is used for backing up Subversion repositories.

    The following restrictions exist on data in this class:

       - The collect mode must be one of the values in :any:`VALID_COLLECT_MODES`.
       - The compress mode must be one of the values in :any:`VALID_COMPRESS_MODES`.
       - The repositories list must be a list of ``Repository`` objects.
       - The repositoryDirs list must be a list of ``RepositoryDir`` objects.

    For the two lists, validation is accomplished through the
    :any:`util.ObjectTypeList` list implementation that overrides common list
    methods and transparently ensures that each element has the correct type.

    *Note:* Lists within this class are "unordered" for equality comparisons.

    """

    def __init__(self, collectMode=None, compressMode=None, repositories=None, repositoryDirs=None):
        """
        Constructor for the ``SubversionConfig`` class.

        Args:
           collectMode: Default collect mode
           compressMode: Default compress mode
           repositories: List of Subversion repositories to back up
           repositoryDirs: List of Subversion parent directories to back up

        Raises:
           ValueError: If one of the values is invalid
        """
        self._collectMode = None
        self._compressMode = None
        self._repositories = None
        self._repositoryDirs = None
        self.collectMode = collectMode
        self.compressMode = compressMode
        self.repositories = repositories
        self.repositoryDirs = repositoryDirs

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "SubversionConfig(%s, %s, %s, %s)" % (self.collectMode, self.compressMode, self.repositories, self.repositoryDirs)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Lists within this class are "unordered" for equality comparisons.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.collectMode != other.collectMode:
            if str(self.collectMode or "") < str(other.collectMode or ""):
                return -1
            else:
                return 1
        if self.compressMode != other.compressMode:
            if str(self.compressMode or "") < str(other.compressMode or ""):
                return -1
            else:
                return 1
        if self.repositories != other.repositories:
            if self.repositories < other.repositories:
                return -1
            else:
                return 1
        if self.repositoryDirs != other.repositoryDirs:
            if self.repositoryDirs < other.repositoryDirs:
                return -1
            else:
                return 1
        return 0

    def _setCollectMode(self, value):
        """
        Property target used to set the collect mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_COLLECT_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_COLLECT_MODES:
                raise ValueError("Collect mode must be one of %s." % VALID_COLLECT_MODES)
        self._collectMode = value

    def _getCollectMode(self):
        """
        Property target used to get the collect mode.
        """
        return self._collectMode

    def _setCompressMode(self, value):
        """
        Property target used to set the compress mode.
        If not ``None``, the mode must be one of the values in :any:`VALID_COMPRESS_MODES`.
        Raises:
           ValueError: If the value is not valid
        """
        if value is not None:
            if value not in VALID_COMPRESS_MODES:
                raise ValueError("Compress mode must be one of %s." % VALID_COMPRESS_MODES)
        self._compressMode = value

    def _getCompressMode(self):
        """
        Property target used to get the compress mode.
        """
        return self._compressMode

    def _setRepositories(self, value):
        """
        Property target used to set the repositories list.
        Either the value must be ``None`` or each element must be a ``Repository``.
        Raises:
           ValueError: If the value is not a ``Repository``
        """
        if value is None:
            self._repositories = None
        else:
            try:
                saved = self._repositories
                self._repositories = ObjectTypeList(Repository, "Repository")
                self._repositories.extend(value)
            except Exception as e:
                self._repositories = saved
                raise e

    def _getRepositories(self):
        """
        Property target used to get the repositories list.
        """
        return self._repositories

    def _setRepositoryDirs(self, value):
        """
        Property target used to set the repositoryDirs list.
        Either the value must be ``None`` or each element must be a ``Repository``.
        Raises:
           ValueError: If the value is not a ``Repository``
        """
        if value is None:
            self._repositoryDirs = None
        else:
            try:
                saved = self._repositoryDirs
                self._repositoryDirs = ObjectTypeList(RepositoryDir, "RepositoryDir")
                self._repositoryDirs.extend(value)
            except Exception as e:
                self._repositoryDirs = saved
                raise e

    def _getRepositoryDirs(self):
        """
        Property target used to get the repositoryDirs list.
        """
        return self._repositoryDirs

    collectMode = property(_getCollectMode, _setCollectMode, None, doc="Default collect mode.")
    compressMode = property(_getCompressMode, _setCompressMode, None, doc="Default compress mode.")
    repositories = property(_getRepositories, _setRepositories, None, doc="List of Subversion repositories to back up.")
    repositoryDirs = property(_getRepositoryDirs, _setRepositoryDirs, None, doc="List of Subversion parent directories to back up.")


########################################################################
# LocalConfig class definition
########################################################################


@total_ordering
class LocalConfig(object):
    """
    Class representing this extension's configuration document.

    This is not a general-purpose configuration object like the main Cedar
    Backup configuration object.  Instead, it just knows how to parse and emit
    Subversion-specific configuration values.  Third parties who need to read
    and write configuration related to this extension should access it through
    the constructor, ``validate`` and ``addConfig`` methods.

    *Note:* Lists within this class are "unordered" for equality comparisons.

    """

    def __init__(self, xmlData=None, xmlPath=None, validate=True):
        """
        Initializes a configuration object.

        If you initialize the object without passing either ``xmlData`` or
        ``xmlPath`` then configuration will be empty and will be invalid until it
        is filled in properly.

        No reference to the original XML data or original path is saved off by
        this class.  Once the data has been parsed (successfully or not) this
        original information is discarded.

        Unless the ``validate`` argument is ``False``, the :any:`LocalConfig.validate`
        method will be called (with its default arguments) against configuration
        after successfully parsing any passed-in XML.  Keep in mind that even if
        ``validate`` is ``False``, it might not be possible to parse the passed-in
        XML document if lower-level validations fail.

        *Note:* It is strongly suggested that the ``validate`` option always be set
        to ``True`` (the default) unless there is a specific need to read in
        invalid configuration from disk.

        Args:
           xmlData (String data): XML data representing configuration
           xmlPath (Absolute path to a file on disk): Path to an XML file on disk
           validate (Boolean true/false): Validate the document after parsing it
        Raises:
           ValueError: If both ``xmlData`` and ``xmlPath`` are passed-in
           ValueError: If the XML data in ``xmlData`` or ``xmlPath`` cannot be parsed
           ValueError: If the parsed configuration document is not valid
        """
        self._subversion = None
        self.subversion = None
        if xmlData is not None and xmlPath is not None:
            raise ValueError("Use either xmlData or xmlPath, but not both.")
        if xmlData is not None:
            self._parseXmlData(xmlData)
            if validate:
                self.validate()
        elif xmlPath is not None:
            with open(xmlPath) as f:  # pylint: disable=unspecified-encoding
                xmlData = f.read()
            self._parseXmlData(xmlData)
            if validate:
                self.validate()

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "LocalConfig(%s)" % (self.subversion)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Lists within this class are "unordered" for equality comparisons.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.subversion != other.subversion:
            if self.subversion < other.subversion:
                return -1
            else:
                return 1
        return 0

    def _setSubversion(self, value):
        """
        Property target used to set the subversion configuration value.
        If not ``None``, the value must be a ``SubversionConfig`` object.
        Raises:
           ValueError: If the value is not a ``SubversionConfig``
        """
        if value is None:
            self._subversion = None
        else:
            if not isinstance(value, SubversionConfig):
                raise ValueError("Value must be a ``SubversionConfig`` object.")
            self._subversion = value

    def _getSubversion(self):
        """
        Property target used to get the subversion configuration value.
        """
        return self._subversion

    subversion = property(
        _getSubversion, _setSubversion, None, "Subversion configuration in terms of a ``SubversionConfig`` object."
    )

    def validate(self):
        """
        Validates configuration represented by the object.

        Subversion configuration must be filled in.  Within that, the collect
        mode and compress mode are both optional, but the list of repositories
        must contain at least one entry.

        Each repository must contain a repository path, and then must be either
        able to take collect mode and compress mode configuration from the parent
        ``SubversionConfig`` object, or must set each value on its own.

        Raises:
           ValueError: If one of the validations fails
        """
        if self.subversion is None:
            raise ValueError("Subversion section is required.")
        if (self.subversion.repositories is None or len(self.subversion.repositories) < 1) and (
            self.subversion.repositoryDirs is None or len(self.subversion.repositoryDirs) < 1
        ):
            raise ValueError("At least one Subversion repository must be configured.")
        if self.subversion.repositories is not None:
            for repository in self.subversion.repositories:
                if repository.repositoryPath is None:
                    raise ValueError("Each repository must set a repository path.")
                if self.subversion.collectMode is None and repository.collectMode is None:
                    raise ValueError("Collect mode must either be set in parent section or individual repository.")
                if self.subversion.compressMode is None and repository.compressMode is None:
                    raise ValueError("Compress mode must either be set in parent section or individual repository.")
        if self.subversion.repositoryDirs is not None:
            for repositoryDir in self.subversion.repositoryDirs:
                if repositoryDir.directoryPath is None:
                    raise ValueError("Each repository directory must set a directory path.")
                if self.subversion.collectMode is None and repositoryDir.collectMode is None:
                    raise ValueError("Collect mode must either be set in parent section or repository directory.")
                if self.subversion.compressMode is None and repositoryDir.compressMode is None:
                    raise ValueError("Compress mode must either be set in parent section or repository directory.")

    def addConfig(self, xmlDom, parentNode):
        """
        Adds a <subversion> configuration section as the next child of a parent.

        Third parties should use this function to write configuration related to
        this extension.

        We add the following fields to the document::

           collectMode    //cb_config/subversion/collectMode
           compressMode   //cb_config/subversion/compressMode

        We also add groups of the following items, one list element per
        item::

           repository     //cb_config/subversion/repository
           repository_dir //cb_config/subversion/repository_dir

        Args:
           xmlDom: DOM tree as from ``impl.createDocument()``
           parentNode: Parent that the section should be appended to
        """
        if self.subversion is not None:
            sectionNode = addContainerNode(xmlDom, parentNode, "subversion")
            addStringNode(xmlDom, sectionNode, "collect_mode", self.subversion.collectMode)
            addStringNode(xmlDom, sectionNode, "compress_mode", self.subversion.compressMode)
            if self.subversion.repositories is not None:
                for repository in self.subversion.repositories:
                    LocalConfig._addRepository(xmlDom, sectionNode, repository)
            if self.subversion.repositoryDirs is not None:
                for repositoryDir in self.subversion.repositoryDirs:
                    LocalConfig._addRepositoryDir(xmlDom, sectionNode, repositoryDir)

    def _parseXmlData(self, xmlData):
        """
        Internal method to parse an XML string into the object.

        This method parses the XML document into a DOM tree (``xmlDom``) and then
        calls a static method to parse the subversion configuration section.

        Args:
           xmlData (String data): XML data to be parsed
        Raises:
           ValueError: If the XML cannot be successfully parsed
        """
        (xmlDom, parentNode) = createInputDom(xmlData)
        self._subversion = LocalConfig._parseSubversion(parentNode)

    @staticmethod
    def _parseSubversion(parent):
        """
        Parses a subversion configuration section.

        We read the following individual fields::

           collectMode    //cb_config/subversion/collect_mode
           compressMode   //cb_config/subversion/compress_mode

        We also read groups of the following item, one list element per
        item::

           repositories    //cb_config/subversion/repository
           repository_dirs //cb_config/subversion/repository_dir

        The repositories are parsed by :any:`_parseRepositories`, and the repository
        dirs are parsed by :any:`_parseRepositoryDirs`.

        Args:
           parent: Parent node to search beneath

        Returns:
            ``SubversionConfig`` object or ``None`` if the section does not exist
        Raises:
           ValueError: If some filled-in value is invalid
        """
        subversion = None
        section = readFirstChild(parent, "subversion")
        if section is not None:
            subversion = SubversionConfig()
            subversion.collectMode = readString(section, "collect_mode")
            subversion.compressMode = readString(section, "compress_mode")
            subversion.repositories = LocalConfig._parseRepositories(section)
            subversion.repositoryDirs = LocalConfig._parseRepositoryDirs(section)
        return subversion

    @staticmethod
    def _parseRepositories(parent):
        """
        Reads a list of ``Repository`` objects from immediately beneath the parent.

        We read the following individual fields::

           repositoryType          type
           repositoryPath          abs_path
           collectMode             collect_mode
           compressMode            compess_mode

        The type field is optional, and its value is kept around only for
        reference.

        Args:
           parent: Parent node to search beneath

        Returns:
            List of ``Repository`` objects or ``None`` if none are found
        Raises:
           ValueError: If some filled-in value is invalid
        """
        lst = []
        for entry in readChildren(parent, "repository"):
            if isElement(entry):
                repository = Repository()
                repository.repositoryType = readString(entry, "type")
                repository.repositoryPath = readString(entry, "abs_path")
                repository.collectMode = readString(entry, "collect_mode")
                repository.compressMode = readString(entry, "compress_mode")
                lst.append(repository)
        if not lst:
            lst = None
        return lst

    @staticmethod
    def _addRepository(xmlDom, parentNode, repository):
        """
        Adds a repository container as the next child of a parent.

        We add the following fields to the document::

           repositoryType          repository/type
           repositoryPath          repository/abs_path
           collectMode             repository/collect_mode
           compressMode            repository/compress_mode

        The <repository> node itself is created as the next child of the parent
        node.  This method only adds one repository node.  The parent must loop
        for each repository in the ``SubversionConfig`` object.

        If ``repository`` is ``None``, this method call will be a no-op.

        Args:
           xmlDom: DOM tree as from ``impl.createDocument()``
           parentNode: Parent that the section should be appended to
           repository: Repository to be added to the document
        """
        if repository is not None:
            sectionNode = addContainerNode(xmlDom, parentNode, "repository")
            addStringNode(xmlDom, sectionNode, "type", repository.repositoryType)
            addStringNode(xmlDom, sectionNode, "abs_path", repository.repositoryPath)
            addStringNode(xmlDom, sectionNode, "collect_mode", repository.collectMode)
            addStringNode(xmlDom, sectionNode, "compress_mode", repository.compressMode)

    @staticmethod
    def _parseRepositoryDirs(parent):
        """
        Reads a list of ``RepositoryDir`` objects from immediately beneath the parent.

        We read the following individual fields::

           repositoryType          type
           directoryPath           abs_path
           collectMode             collect_mode
           compressMode            compess_mode

        We also read groups of the following items, one list element per
        item::

           relativeExcludePaths    exclude/rel_path
           excludePatterns         exclude/pattern

        The exclusions are parsed by :any:`_parseExclusions`.

        The type field is optional, and its value is kept around only for
        reference.

        Args:
           parent: Parent node to search beneath

        Returns:
            List of ``RepositoryDir`` objects or ``None`` if none are found
        Raises:
           ValueError: If some filled-in value is invalid
        """
        lst = []
        for entry in readChildren(parent, "repository_dir"):
            if isElement(entry):
                repositoryDir = RepositoryDir()
                repositoryDir.repositoryType = readString(entry, "type")
                repositoryDir.directoryPath = readString(entry, "abs_path")
                repositoryDir.collectMode = readString(entry, "collect_mode")
                repositoryDir.compressMode = readString(entry, "compress_mode")
                (repositoryDir.relativeExcludePaths, repositoryDir.excludePatterns) = LocalConfig._parseExclusions(entry)
                lst.append(repositoryDir)
        if not lst:
            lst = None
        return lst

    @staticmethod
    def _parseExclusions(parentNode):
        """
        Reads exclusions data from immediately beneath the parent.

        We read groups of the following items, one list element per item::

           relative    exclude/rel_path
           patterns    exclude/pattern

        If there are none of some pattern (i.e. no relative path items) then
        ``None`` will be returned for that item in the tuple.

        Args:
           parentNode: Parent node to search beneath

        Returns:
            Tuple of (relative, patterns) exclusions
        """
        section = readFirstChild(parentNode, "exclude")
        if section is None:
            return (None, None)
        else:
            relative = readStringList(section, "rel_path")
            patterns = readStringList(section, "pattern")
            return (relative, patterns)

    @staticmethod
    def _addRepositoryDir(xmlDom, parentNode, repositoryDir):
        """
        Adds a repository dir container as the next child of a parent.

        We add the following fields to the document::

           repositoryType          repository_dir/type
           directoryPath           repository_dir/abs_path
           collectMode             repository_dir/collect_mode
           compressMode            repository_dir/compress_mode

        We also add groups of the following items, one list element per item::

           relativeExcludePaths    dir/exclude/rel_path
           excludePatterns         dir/exclude/pattern

        The <repository_dir> node itself is created as the next child of the
        parent node.  This method only adds one repository node.  The parent must
        loop for each repository dir in the ``SubversionConfig`` object.

        If ``repositoryDir`` is ``None``, this method call will be a no-op.

        Args:
           xmlDom: DOM tree as from ``impl.createDocument()``
           parentNode: Parent that the section should be appended to
           repositoryDir: Repository dir to be added to the document
        """
        if repositoryDir is not None:
            sectionNode = addContainerNode(xmlDom, parentNode, "repository_dir")
            addStringNode(xmlDom, sectionNode, "type", repositoryDir.repositoryType)
            addStringNode(xmlDom, sectionNode, "abs_path", repositoryDir.directoryPath)
            addStringNode(xmlDom, sectionNode, "collect_mode", repositoryDir.collectMode)
            addStringNode(xmlDom, sectionNode, "compress_mode", repositoryDir.compressMode)
            if (repositoryDir.relativeExcludePaths is not None and repositoryDir.relativeExcludePaths != []) or (
                repositoryDir.excludePatterns is not None and repositoryDir.excludePatterns != []
            ):
                excludeNode = addContainerNode(xmlDom, sectionNode, "exclude")
                if repositoryDir.relativeExcludePaths is not None:
                    for relativePath in repositoryDir.relativeExcludePaths:
                        addStringNode(xmlDom, excludeNode, "rel_path", relativePath)
                if repositoryDir.excludePatterns is not None:
                    for pattern in repositoryDir.excludePatterns:
                        addStringNode(xmlDom, excludeNode, "pattern", pattern)


########################################################################
# Public functions
########################################################################

###########################
# executeAction() function
###########################


def executeAction(configPath, options, config):
    """
    Executes the Subversion backup action.

    Args:
       configPath (String representing a path on disk): Path to configuration file on disk
       options (Options object): Program command-line options
       config (Config object): Program configuration
    Raises:
       ValueError: Under many generic error conditions
       IOError: If a backup could not be written for some reason
    """
    logger.debug("Executing Subversion extended action.")
    if config.options is None or config.collect is None:
        raise ValueError("Cedar Backup configuration is not properly filled in.")
    local = LocalConfig(xmlPath=configPath)
    todayIsStart = isStartOfWeek(config.options.startingDay)
    fullBackup = options.full or todayIsStart
    logger.debug("Full backup flag is [%s]", fullBackup)
    if local.subversion.repositories is not None:
        for repository in local.subversion.repositories:
            _backupRepository(config, local, todayIsStart, fullBackup, repository)
    if local.subversion.repositoryDirs is not None:
        for repositoryDir in local.subversion.repositoryDirs:
            logger.debug("Working with repository directory [%s].", repositoryDir.directoryPath)
            for repositoryPath in _getRepositoryPaths(repositoryDir):
                repository = Repository(
                    repositoryDir.repositoryType, repositoryPath, repositoryDir.collectMode, repositoryDir.compressMode
                )
                _backupRepository(config, local, todayIsStart, fullBackup, repository)
            logger.info("Completed backing up Subversion repository directory [%s].", repositoryDir.directoryPath)
    logger.info("Executed the Subversion extended action successfully.")


def _getCollectMode(local, repository):
    """
    Gets the collect mode that should be used for a repository.
    Use repository's if possible, otherwise take from subversion section.
    Args:
       repository: Repository object
    Returns:
        Collect mode to use
    """
    if repository.collectMode is None:
        collectMode = local.subversion.collectMode
    else:
        collectMode = repository.collectMode
    logger.debug("Collect mode is [%s]", collectMode)
    return collectMode


def _getCompressMode(local, repository):
    """
    Gets the compress mode that should be used for a repository.
    Use repository's if possible, otherwise take from subversion section.
    Args:
       local: LocalConfig object
       repository: Repository object
    Returns:
        Compress mode to use
    """
    if repository.compressMode is None:
        compressMode = local.subversion.compressMode
    else:
        compressMode = repository.compressMode
    logger.debug("Compress mode is [%s]", compressMode)
    return compressMode


def _getRevisionPath(config, repository):
    """
    Gets the path to the revision file associated with a repository.
    Args:
       config: Config object
       repository: Repository object
    Returns:
        Absolute path to the revision file associated with the repository
    """
    normalized = buildNormalizedPath(repository.repositoryPath)
    filename = "%s.%s" % (normalized, REVISION_PATH_EXTENSION)
    revisionPath = pathJoin(config.options.workingDir, filename)
    logger.debug("Revision file path is [%s]", revisionPath)
    return revisionPath


def _getBackupPath(config, repositoryPath, compressMode, startRevision, endRevision):
    """
    Gets the backup file path (including correct extension) associated with a repository.
    Args:
       config: Config object
       repositoryPath: Path to the indicated repository
       compressMode: Compress mode to use for this repository
       startRevision: Starting repository revision
       endRevision: Ending repository revision
    Returns:
        Absolute path to the backup file associated with the repository
    """
    normalizedPath = buildNormalizedPath(repositoryPath)
    filename = "svndump-%d:%d-%s.txt" % (startRevision, endRevision, normalizedPath)
    if compressMode == "gzip":
        filename = "%s.gz" % filename
    elif compressMode == "bzip2":
        filename = "%s.bz2" % filename
    backupPath = pathJoin(config.collect.targetDir, filename)
    logger.debug("Backup file path is [%s]", backupPath)
    return backupPath


def _getRepositoryPaths(repositoryDir):
    """
    Gets a list of child repository paths within a repository directory.
    Args:
       repositoryDir: RepositoryDirectory
    """
    (excludePaths, excludePatterns) = _getExclusions(repositoryDir)
    fsList = FilesystemList()
    fsList.excludeFiles = True
    fsList.excludeLinks = True
    fsList.excludePaths = excludePaths
    fsList.excludePatterns = excludePatterns
    fsList.addDirContents(path=repositoryDir.directoryPath, recursive=False, addSelf=False)
    return fsList


def _getExclusions(repositoryDir):
    """
    Gets exclusions (file and patterns) associated with an repository directory.

    The returned files value is a list of absolute paths to be excluded from the
    backup for a given directory.  It is derived from the repository directory's
    relative exclude paths.

    The returned patterns value is a list of patterns to be excluded from the
    backup for a given directory.  It is derived from the repository directory's
    list of patterns.

    Args:
       repositoryDir: Repository directory object

    Returns:
        Tuple (files, patterns) indicating what to exclude
    """
    paths = []
    if repositoryDir.relativeExcludePaths is not None:
        for relativePath in repositoryDir.relativeExcludePaths:
            paths.append(pathJoin(repositoryDir.directoryPath, relativePath))
    patterns = []
    if repositoryDir.excludePatterns is not None:
        patterns.extend(repositoryDir.excludePatterns)
    logger.debug("Exclude paths: %s", paths)
    logger.debug("Exclude patterns: %s", patterns)
    return (paths, patterns)


def _backupRepository(config, local, todayIsStart, fullBackup, repository):
    """
    Backs up an individual Subversion repository.

    This internal method wraps the public methods and adds some functionality
    to work better with the extended action itself.

    Args:
       config: Cedar Backup configuration
       local: Local configuration
       todayIsStart: Indicates whether today is start of week
       fullBackup: Full backup flag
       repository: Repository to operate on

    Raises:
       ValueError: If some value is missing or invalid
       IOError: If there is a problem executing the Subversion dump
    """
    logger.debug("Working with repository [%s]", repository.repositoryPath)
    logger.debug("Repository type is [%s]", repository.repositoryType)
    collectMode = _getCollectMode(local, repository)
    compressMode = _getCompressMode(local, repository)
    revisionPath = _getRevisionPath(config, repository)
    if not (fullBackup or (collectMode in ["daily", "incr"]) or (collectMode == "weekly" and todayIsStart)):
        logger.debug("Repository will not be backed up, per collect mode.")
        return
    logger.debug("Repository meets criteria to be backed up today.")
    if collectMode != "incr" or fullBackup:
        startRevision = 0
        endRevision = getYoungestRevision(repository.repositoryPath)
        logger.debug("Using full backup, revision: (%d, %d).", startRevision, endRevision)
    else:
        if fullBackup:
            startRevision = 0
            endRevision = getYoungestRevision(repository.repositoryPath)
        else:
            startRevision = _loadLastRevision(revisionPath) + 1
            endRevision = getYoungestRevision(repository.repositoryPath)
            if startRevision > endRevision:
                logger.info("No need to back up repository [%s]; no new revisions.", repository.repositoryPath)
                return
        logger.debug("Using incremental backup, revision: (%d, %d).", startRevision, endRevision)
    backupPath = _getBackupPath(config, repository.repositoryPath, compressMode, startRevision, endRevision)
    with _getOutputFile(backupPath, compressMode) as outputFile:
        backupRepository(repository.repositoryPath, outputFile, startRevision, endRevision)
    if not os.path.exists(backupPath):
        raise IOError("Dump file [%s] does not seem to exist after backup completed." % backupPath)
    changeOwnership(backupPath, config.options.backupUser, config.options.backupGroup)
    if collectMode == "incr":
        _writeLastRevision(config, revisionPath, endRevision)
    logger.info("Completed backing up Subversion repository [%s].", repository.repositoryPath)


def _getOutputFile(backupPath, compressMode):
    """
    Opens the output file used for saving the Subversion dump.

    If the compress mode is "gzip", we'll open a ``GzipFile``, and if the
    compress mode is "bzip2", we'll open a ``BZ2File``.  Otherwise, we'll just
    return an object from the normal ``open()`` method.

    Args:
       backupPath: Path to file to open
       compressMode: Compress mode of file ("none", "gzip", "bzip")

    Returns:
        Output file object, opened in binary mode for use with executeCommand()
    """
    if compressMode == "gzip":
        return GzipFile(backupPath, "wb")
    elif compressMode == "bzip2":
        return BZ2File(backupPath, "wb")
    else:
        return open(backupPath, "wb")


def _loadLastRevision(revisionPath):
    """
    Loads the indicated revision file from disk into an integer.

    If we can't load the revision file successfully (either because it doesn't
    exist or for some other reason), then a revision of -1 will be returned -
    but the condition will be logged.  This way, we err on the side of backing
    up too much, because anyone using this will presumably be adding 1 to the
    revision, so they don't duplicate any backups.

    Args:
       revisionPath: Path to the revision file on disk

    Returns:
        Integer representing last backed-up revision, -1 on error or if none can be read
    """
    if not os.path.isfile(revisionPath):
        startRevision = -1
        logger.debug("Revision file [%s] does not exist on disk.", revisionPath)
    else:
        try:
            with open(revisionPath, "rb") as f:
                startRevision = pickle.load(f, fix_imports=True)  # be compatible with Python 2
            logger.debug("Loaded revision file [%s] from disk: %d.", revisionPath, startRevision)
        except Exception as e:
            startRevision = -1
            logger.error("Failed loading revision file [%s] from disk: %s", revisionPath, e)
    return startRevision


def _writeLastRevision(config, revisionPath, endRevision):
    """
    Writes the end revision to the indicated revision file on disk.

    If we can't write the revision file successfully for any reason, we'll log
    the condition but won't throw an exception.

    Args:
       config: Config object
       revisionPath: Path to the revision file on disk
       endRevision: Last revision backed up on this run
    """
    try:
        with open(revisionPath, "wb") as f:
            pickle.dump(endRevision, f, 0, fix_imports=True)
        changeOwnership(revisionPath, config.options.backupUser, config.options.backupGroup)
        logger.debug("Wrote new revision file [%s] to disk: %d.", revisionPath, endRevision)
    except Exception as e:
        logger.error("Failed to write revision file [%s] to disk: %s", revisionPath, e)


##############################
# backupRepository() function
##############################


def backupRepository(repositoryPath, backupFile, startRevision=None, endRevision=None):
    """
    Backs up an individual Subversion repository.

    The starting and ending revision values control an incremental backup.  If
    the starting revision is not passed in, then revision zero (the start of the
    repository) is assumed.  If the ending revision is not passed in, then the
    youngest revision in the database will be used as the endpoint.

    The backup data will be written into the passed-in back file.  Normally,
    this would be an object as returned from ``open``, but it is possible to use
    something like a ``GzipFile`` to write compressed output.  The caller is
    responsible for closing the passed-in backup file.

    *Note:* This function should either be run as root or as the owner of the
    Subversion repository.

    *Note:* It is apparently *not* a good idea to interrupt this function.
    Sometimes, this leaves the repository in a "wedged" state, which requires
    recovery using ``svnadmin recover``.

    Args:
       repositoryPath (String path representing Subversion repository on disk): Path to Subversion repository to back up
       backupFile (Python file object as from ``open`` or ``file``): Python file object to use for writing backup
       startRevision (Integer value >= 0): Starting repository revision to back up (for incremental backups)
       endRevision (Integer value >= 0): Ending repository revision to back up (for incremental backups)
    Raises:
       ValueError: If some value is missing or invalid
       IOError: If there is a problem executing the Subversion dump
    """
    if startRevision is None:
        startRevision = 0
    if endRevision is None:
        endRevision = getYoungestRevision(repositoryPath)
    if int(startRevision) < 0:
        raise ValueError("Start revision must be >= 0.")
    if int(endRevision) < 0:
        raise ValueError("End revision must be >= 0.")
    if startRevision > endRevision:
        raise ValueError("Start revision must be <= end revision.")
    args = ["dump", "--quiet", "-r%s:%s" % (startRevision, endRevision), "--incremental", repositoryPath]
    command = resolveCommand(SVNADMIN_COMMAND)
    result = executeCommand(command, args, returnOutput=False, ignoreStderr=True, doNotLog=True, outputFile=backupFile)[0]
    if result != 0:
        raise IOError("Error [%d] executing Subversion dump for repository [%s]." % (result, repositoryPath))
    logger.debug("Completed dumping subversion repository [%s].", repositoryPath)


#################################
# getYoungestRevision() function
#################################


def getYoungestRevision(repositoryPath):
    """
    Gets the youngest (newest) revision in a Subversion repository using ``svnlook``.

    *Note:* This function should either be run as root or as the owner of the
    Subversion repository.

    Args:
       repositoryPath (String path representing Subversion repository on disk): Path to Subversion repository to look in
    Returns:
        Youngest revision as an integer

    Raises:
       ValueError: If there is a problem parsing the ``svnlook`` output
       IOError: If there is a problem executing the ``svnlook`` command
    """
    args = ["youngest", repositoryPath]
    command = resolveCommand(SVNLOOK_COMMAND)
    (result, output) = executeCommand(command, args, returnOutput=True, ignoreStderr=True)
    if result != 0:
        raise IOError("Error [%d] executing 'svnlook youngest' for repository [%s]." % (result, repositoryPath))
    if len(output) != 1:
        raise ValueError("Unable to parse 'svnlook youngest' output.")
    return int(output[0])


########################################################################
# Deprecated functionality
########################################################################


class BDBRepository(Repository):
    """
    Class representing Subversion BDB (Berkeley Database) repository configuration.
    This object is deprecated.  Use a simple :any:`Repository` instead.
    """

    def __init__(self, repositoryPath=None, collectMode=None, compressMode=None):
        """
        Constructor for the ``BDBRepository`` class.
        """
        super(BDBRepository, self).__init__("BDB", repositoryPath, collectMode, compressMode)

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "BDBRepository(%s, %s, %s)" % (self.repositoryPath, self.collectMode, self.compressMode)


class FSFSRepository(Repository):
    """
    Class representing Subversion FSFS repository configuration.
    This object is deprecated.  Use a simple :any:`Repository` instead.
    """

    def __init__(self, repositoryPath=None, collectMode=None, compressMode=None):
        """
        Constructor for the ``FSFSRepository`` class.
        """
        super(FSFSRepository, self).__init__("FSFS", repositoryPath, collectMode, compressMode)

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "FSFSRepository(%s, %s, %s)" % (self.repositoryPath, self.collectMode, self.compressMode)


def backupBDBRepository(repositoryPath, backupFile, startRevision=None, endRevision=None):
    """
    Backs up an individual Subversion BDB repository.
    This function is deprecated.  Use :any:`backupRepository` instead.
    """
    return backupRepository(repositoryPath, backupFile, startRevision, endRevision)


def backupFSFSRepository(repositoryPath, backupFile, startRevision=None, endRevision=None):
    """
    Backs up an individual Subversion FSFS repository.
    This function is deprecated.  Use :any:`backupRepository` instead.
    """
    return backupRepository(repositoryPath, backupFile, startRevision, endRevision)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2005,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Official Cedar Backup Extensions
# Purpose  : Provides an extension to save off important system recovery information.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Provides an extension to save off important system recovery information.

This is a simple Cedar Backup extension used to save off important system
recovery information.  It saves off three types of information:

   - Currently-installed Debian packages via ``dpkg --get-selections``
   - Disk partition information via ``fdisk -l``
   - System-wide mounted filesystem contents, via ``ls -laR``

The saved-off information is placed into the collect directory and is
compressed using ``bzip2`` to save space.

This extension relies on the options and collect configurations in the standard
Cedar Backup configuration file, but requires no new configuration of its own.
No public functions other than the action are exposed since all of this is
pretty simple.

*Note:* If the ``dpkg`` or ``fdisk`` commands cannot be found in their normal
locations or executed by the current user, those steps will be skipped and a
note will be logged at the INFO level.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules
########################################################################

import logging
import os
from bz2 import BZ2File

from CedarBackup3.util import changeOwnership, executeCommand, pathJoin, resolveCommand

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.extend.sysinfo")

DPKG_PATH = "/usr/bin/dpkg"
FDISK_PATH = "/sbin/fdisk"

DPKG_COMMAND = [DPKG_PATH, "--get-selections"]
FDISK_COMMAND = [FDISK_PATH, "-l"]
LS_COMMAND = ["ls", "-laR", "/"]


########################################################################
# Public functions
########################################################################

###########################
# executeAction() function
###########################


# pylint: disable=W0613
def executeAction(configPath, options, config):
    """
    Executes the sysinfo backup action.

    Args:
       configPath (String representing a path on disk): Path to configuration file on disk
       options (Options object): Program command-line options
       config (Config object): Program configuration
    Raises:
       ValueError: Under many generic error conditions
       IOError: If the backup process fails for some reason
    """
    logger.debug("Executing sysinfo extended action.")
    if config.options is None or config.collect is None:
        raise ValueError("Cedar Backup configuration is not properly filled in.")
    _dumpDebianPackages(config.collect.targetDir, config.options.backupUser, config.options.backupGroup)
    _dumpPartitionTable(config.collect.targetDir, config.options.backupUser, config.options.backupGroup)
    _dumpFilesystemContents(config.collect.targetDir, config.options.backupUser, config.options.backupGroup)
    logger.info("Executed the sysinfo extended action successfully.")


def _dumpDebianPackages(targetDir, backupUser, backupGroup, compress=True):
    """
    Dumps a list of currently installed Debian packages via ``dpkg``.
    Args:
       targetDir: Directory to write output file into
       backupUser: User which should own the resulting file
       backupGroup: Group which should own the resulting file
       compress: Indicates whether to compress the output file
    Raises:
       IOError: If the dump fails for some reason
    """
    if not os.path.exists(DPKG_PATH):
        logger.info("Not executing Debian package dump since %s doesn't seem to exist.", DPKG_PATH)
    elif not os.access(DPKG_PATH, os.X_OK):
        logger.info("Not executing Debian package dump since %s cannot be executed.", DPKG_PATH)
    else:
        (outputFile, filename) = _getOutputFile(targetDir, "dpkg-selections", compress)
        with outputFile:
            command = resolveCommand(DPKG_COMMAND)
            result = executeCommand(command, [], returnOutput=False, ignoreStderr=True, doNotLog=True, outputFile=outputFile)[0]
            if result != 0:
                raise IOError("Error [%d] executing Debian package dump." % result)
        if not os.path.exists(filename):
            raise IOError("File [%s] does not seem to exist after Debian package dump finished." % filename)
        changeOwnership(filename, backupUser, backupGroup)


def _dumpPartitionTable(targetDir, backupUser, backupGroup, compress=True):
    """
    Dumps information about the partition table via ``fdisk``.
    Args:
       targetDir: Directory to write output file into
       backupUser: User which should own the resulting file
       backupGroup: Group which should own the resulting file
       compress: Indicates whether to compress the output file
    Raises:
       IOError: If the dump fails for some reason
    """
    if not os.path.exists(FDISK_PATH):
        logger.info("Not executing partition table dump since %s doesn't seem to exist.", FDISK_PATH)
    elif not os.access(FDISK_PATH, os.X_OK):
        logger.info("Not executing partition table dump since %s cannot be executed.", FDISK_PATH)
    else:
        (outputFile, filename) = _getOutputFile(targetDir, "fdisk-l", compress)
        with outputFile:
            command = resolveCommand(FDISK_COMMAND)
            result = executeCommand(command, [], returnOutput=False, ignoreStderr=True, outputFile=outputFile)[0]
            if result != 0:
                raise IOError("Error [%d] executing partition table dump." % result)
        if not os.path.exists(filename):
            raise IOError("File [%s] does not seem to exist after partition table dump finished." % filename)
        changeOwnership(filename, backupUser, backupGroup)


def _dumpFilesystemContents(targetDir, backupUser, backupGroup, compress=True):
    """
    Dumps complete listing of filesystem contents via ``ls -laR``.
    Args:
       targetDir: Directory to write output file into
       backupUser: User which should own the resulting file
       backupGroup: Group which should own the resulting file
       compress: Indicates whether to compress the output file
    Raises:
       IOError: If the dump fails for some reason
    """
    (outputFile, filename) = _getOutputFile(targetDir, "ls-laR", compress)
    with outputFile:
        # Note: can't count on return status from 'ls', so we don't check it.
        command = resolveCommand(LS_COMMAND)
        executeCommand(command, [], returnOutput=False, ignoreStderr=True, doNotLog=True, outputFile=outputFile)
    if not os.path.exists(filename):
        raise IOError("File [%s] does not seem to exist after filesystem contents dump finished." % filename)
    changeOwnership(filename, backupUser, backupGroup)


def _getOutputFile(targetDir, name, compress=True):
    """
    Opens the output file used for saving a dump to the filesystem.

    The filename will be ``name.txt`` (or ``name.txt.bz2`` if ``compress`` is
    ``True``), written in the target directory.

    Args:
       targetDir: Target directory to write file in
       name: Name of the file to create
       compress: Indicates whether to write compressed output

    Returns:
        Tuple of (Output file object, filename), file opened in binary mode for use with executeCommand()
    """
    filename = pathJoin(targetDir, "%s.txt" % name)
    if compress:
        filename = "%s.bz2" % filename
    logger.debug("Dump file will be [%s].", filename)
    if compress:
        outputFile = BZ2File(filename, "wb")
    else:
        outputFile = open(filename, "wb")
    return (outputFile, filename)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2004-2007,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Implements the standard 'rebuild' action.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Implements the standard 'rebuild' action.
:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""


########################################################################
# Imported modules
########################################################################

import datetime
import logging
import os
import sys

from CedarBackup3.actions.constants import DIR_TIME_FORMAT, STAGE_INDICATOR
from CedarBackup3.actions.store import consistencyCheck, writeImage, writeStoreIndicator
from CedarBackup3.actions.util import checkMediaState
from CedarBackup3.util import deriveDayOfWeek, pathJoin

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.actions.rebuild")


########################################################################
# Public functions
########################################################################

############################
# executeRebuild() function
############################


# pylint: disable=W0613
def executeRebuild(configPath, options, config):
    """
    Executes the rebuild backup action.

    This function exists mainly to recreate a disc that has been "trashed" due
    to media or hardware problems.  Note that the "stage complete" indicator
    isn't checked for this action.

    Note that the rebuild action and the store action are very similar.  The
    main difference is that while store only stores a single day's staging
    directory, the rebuild action operates on multiple staging directories.

    Args:
       configPath (String representing a path on disk): Path to configuration file on disk
       options (Options object): Program command-line options
       config (Config object): Program configuration
    Raises:
       ValueError: Under many generic error conditions
       IOError: If there are problems reading or writing files
    """
    logger.debug("Executing the 'rebuild' action.")
    if sys.platform == "darwin":
        logger.warning("Warning: the rebuild action is not fully supported on Mac OS X.")
        logger.warning("See the Cedar Backup software manual for further information.")
    if config.options is None or config.store is None:
        raise ValueError("Rebuild configuration is not properly filled in.")
    if config.store.checkMedia:
        checkMediaState(config.store)  # raises exception if media is not initialized
    stagingDirs = _findRebuildDirs(config)
    writeImage(config, True, stagingDirs)
    if config.store.checkData:
        if sys.platform == "darwin":
            logger.warning("Warning: consistency check cannot be run successfully on Mac OS X.")
            logger.warning("See the Cedar Backup software manual for further information.")
        else:
            logger.debug("Running consistency check of media.")
            consistencyCheck(config, stagingDirs)
    writeStoreIndicator(config, stagingDirs)
    logger.info("Executed the 'rebuild' action successfully.")


########################################################################
# Private utility functions
########################################################################

##############################
# _findRebuildDirs() function
##############################


def _findRebuildDirs(config):
    """
    Finds the set of directories to be included in a disc rebuild.

    A the rebuild action is supposed to recreate the "last week's" disc.  This
    won't always be possible if some of the staging directories are missing.
    However, the general procedure is to look back into the past no further than
    the previous "starting day of week", and then work forward from there trying
    to find all of the staging directories between then and now that still exist
    and have a stage indicator.

    Args:
       config: Config object

    Returns:
        Correct staging dir, as a dict mapping directory to date suffix
    Raises:
       IOError: If we do not find at least one staging directory
    """
    stagingDirs = {}
    start = deriveDayOfWeek(config.options.startingDay)
    today = datetime.date.today()
    if today.weekday() >= start:
        days = today.weekday() - start + 1
    else:
        days = 7 - (start - today.weekday()) + 1
    for i in range(0, days):
        currentDay = today - datetime.timedelta(days=i)
        dateSuffix = currentDay.strftime(DIR_TIME_FORMAT)
        stageDir = pathJoin(config.store.sourceDir, dateSuffix)
        indicator = pathJoin(stageDir, STAGE_INDICATOR)
        if os.path.isdir(stageDir) and os.path.exists(indicator):
            logger.info("Rebuild process will include stage directory [%s]", stageDir)
            stagingDirs[stageDir] = dateSuffix
    if len(stagingDirs) == 0:
        raise IOError("Unable to find any staging directories for rebuild process.")
    return stagingDirs



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Implements customized behavior.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Implements customized behavior.

Some behaviors need to vary when packaged for certain platforms.  For instance,
while Cedar Backup generally uses cdrecord and mkisofs, Debian ships compatible
utilities called wodim and genisoimage. I want there to be one single place
where Cedar Backup is patched for Debian, rather than having to maintain a
variety of patches in different places.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules
########################################################################

import logging

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.customize")

PLATFORM = "standard"
# PLATFORM = "debian"

DEBIAN_CDRECORD = "/usr/bin/wodim"
DEBIAN_MKISOFS = "/usr/bin/genisoimage"


#######################################################################
# Public functions
#######################################################################

################################
# customizeOverrides() function
################################


def customizeOverrides(config, platform=PLATFORM):
    """
    Modify command overrides based on the configured platform.

    On some platforms, we want to add command overrides to configuration.  Each
    override will only be added if the configuration does not already contain an
    override with the same name.  That way, the user still has a way to choose
    their own version of the command if they want.

    Args:
       config: Configuration to modify
       platform: Platform that is in use
    """
    if platform == "debian":
        logger.info("Overriding cdrecord for Debian platform: %s", DEBIAN_CDRECORD)
        config.options.addOverride("cdrecord", DEBIAN_CDRECORD)
        logger.info("Overriding mkisofs for Debian platform: %s", DEBIAN_MKISOFS)
        config.options.addOverride("mkisofs", DEBIAN_MKISOFS)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Official Cedar Backup Tools
# Purpose  : Provides package initialization
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Official Cedar Backup Tools

This package provides official Cedar Backup tools.  Tools are things that feel
a little like extensions, but don't fit the normal mold of extensions.  For
instance, they might not be intended to run from cron, or might need to interact
dynamically with the user (i.e. accept user input).

Tools are usually scripts that are run directly from the command line, just
like the main ``cback3`` script.  Like the ``cback3`` script, the majority of a
tool is implemented in a .py module, and then the script just invokes the
module's ``cli()`` function.  The actual scripts for tools are distributed in
the util/ directory.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""


########################################################################
# Package initialization
########################################################################

# Using 'from CedarBackup3.tools import *' will just import the modules listed
# in the __all__ variable.

import CedarBackup3.tools.amazons3
import CedarBackup3.tools.span

__all__ = ["span", "amazons3"]



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2008,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Provides an extension to check remaining media capacity.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Provides an extension to check remaining media capacity.

Some users have asked for advance warning that their media is beginning to fill
up.  This is an extension that checks the current capacity of the media in the
writer, and prints a warning if the media is more than X% full, or has fewer
than X bytes of capacity remaining.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules
########################################################################

import logging
from functools import total_ordering

from CedarBackup3.actions.util import checkMediaState, createWriter
from CedarBackup3.config import ByteQuantity, addByteQuantityNode, readByteQuantity
from CedarBackup3.util import displayBytes
from CedarBackup3.xmlutil import addContainerNode, addStringNode, createInputDom, readFirstChild, readString

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.extend.capacity")


########################################################################
# Percentage class definition
########################################################################


@total_ordering
class PercentageQuantity(object):
    """
    Class representing a percentage quantity.

    The percentage is maintained internally as a string so that issues of
    precision can be avoided.  It really isn't possible to store a floating
    point number here while being able to losslessly translate back and forth
    between XML and object representations.  (Perhaps the Python 2.4 Decimal
    class would have been an option, but I originally wanted to stay compatible
    with Python 2.3.)

    Even though the quantity is maintained as a string, the string must be in a
    valid floating point positive number.  Technically, any floating point
    string format supported by Python is allowble.  However, it does not make
    sense to have a negative percentage in this context.

    """

    def __init__(self, quantity=None):
        """
        Constructor for the ``PercentageQuantity`` class.
        Args:
           quantity: Percentage quantity, as a string (i.e. "99.9" or "12")
        Raises:
           ValueError: If the quantity value is invaid
        """
        self._quantity = None
        self.quantity = quantity

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "PercentageQuantity(%s)" % (self.quantity)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Lists within this class are "unordered" for equality comparisons.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.quantity != other.quantity:
            if float(self.quantity or 0.0) < float(other.quantity or 0.0):
                return -1
            else:
                return 1
        return 0

    def _setQuantity(self, value):
        """
        Property target used to set the quantity
        The value must be a non-empty string if it is not ``None``.
        Raises:
           ValueError: If the value is an empty string
           ValueError: If the value is not a valid floating point number
           ValueError: If the value is less than zero
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("Percentage must be a non-empty string.")
            floatValue = float(value)
            if floatValue < 0.0 or floatValue > 100.0:
                raise ValueError("Percentage must be a positive value from 0.0 to 100.0")
        self._quantity = value  # keep around string

    def _getQuantity(self):
        """
        Property target used to get the quantity.
        """
        return self._quantity

    def _getPercentage(self):
        """
        Property target used to get the quantity as a floating point number.
        If there is no quantity set, then a value of 0.0 is returned.
        """
        if self.quantity is not None:
            return float(self.quantity)
        return 0.0

    quantity = property(_getQuantity, _setQuantity, None, doc="Percentage value, as a string")
    percentage = property(_getPercentage, None, None, "Percentage value, as a floating point number.")


########################################################################
# CapacityConfig class definition
########################################################################


@total_ordering
class CapacityConfig(object):
    """
    Class representing capacity configuration.

    The following restrictions exist on data in this class:

       - The maximum percentage utilized must be a PercentageQuantity
       - The minimum bytes remaining must be a ByteQuantity

    """

    def __init__(self, maxPercentage=None, minBytes=None):
        """
        Constructor for the ``CapacityConfig`` class.

        Args:
           maxPercentage: Maximum percentage of the media that may be utilized
           minBytes: Minimum number of free bytes that must be available
        """
        self._maxPercentage = None
        self._minBytes = None
        self.maxPercentage = maxPercentage
        self.minBytes = minBytes

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "CapacityConfig(%s, %s)" % (self.maxPercentage, self.minBytes)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.maxPercentage != other.maxPercentage:
            if (self.maxPercentage or PercentageQuantity()) < (other.maxPercentage or PercentageQuantity()):
                return -1
            else:
                return 1
        if self.minBytes != other.minBytes:
            if (self.minBytes or ByteQuantity()) < (other.minBytes or ByteQuantity()):
                return -1
            else:
                return 1
        return 0

    def _setMaxPercentage(self, value):
        """
        Property target used to set the maxPercentage value.
        If not ``None``, the value must be a ``PercentageQuantity`` object.
        Raises:
           ValueError: If the value is not a ``PercentageQuantity``
        """
        if value is None:
            self._maxPercentage = None
        else:
            if not isinstance(value, PercentageQuantity):
                raise ValueError("Value must be a ``PercentageQuantity`` object.")
            self._maxPercentage = value

    def _getMaxPercentage(self):
        """
        Property target used to get the maxPercentage value
        """
        return self._maxPercentage

    def _setMinBytes(self, value):
        """
        Property target used to set the bytes utilized value.
        If not ``None``, the value must be a ``ByteQuantity`` object.
        Raises:
           ValueError: If the value is not a ``ByteQuantity``
        """
        if value is None:
            self._minBytes = None
        else:
            if not isinstance(value, ByteQuantity):
                raise ValueError("Value must be a ``ByteQuantity`` object.")
            self._minBytes = value

    def _getMinBytes(self):
        """
        Property target used to get the bytes remaining value.
        """
        return self._minBytes

    maxPercentage = property(_getMaxPercentage, _setMaxPercentage, None, "Maximum percentage of the media that may be utilized.")
    minBytes = property(_getMinBytes, _setMinBytes, None, "Minimum number of free bytes that must be available.")


########################################################################
# LocalConfig class definition
########################################################################


@total_ordering
class LocalConfig(object):
    """
    Class representing this extension's configuration document.

    This is not a general-purpose configuration object like the main Cedar
    Backup configuration object.  Instead, it just knows how to parse and emit
    specific configuration values to this extension.  Third parties who need to
    read and write configuration related to this extension should access it
    through the constructor, ``validate`` and ``addConfig`` methods.

    *Note:* Lists within this class are "unordered" for equality comparisons.

    """

    def __init__(self, xmlData=None, xmlPath=None, validate=True):
        """
        Initializes a configuration object.

        If you initialize the object without passing either ``xmlData`` or
        ``xmlPath`` then configuration will be empty and will be invalid until it
        is filled in properly.

        No reference to the original XML data or original path is saved off by
        this class.  Once the data has been parsed (successfully or not) this
        original information is discarded.

        Unless the ``validate`` argument is ``False``, the :any:`LocalConfig.validate`
        method will be called (with its default arguments) against configuration
        after successfully parsing any passed-in XML.  Keep in mind that even if
        ``validate`` is ``False``, it might not be possible to parse the passed-in
        XML document if lower-level validations fail.

        *Note:* It is strongly suggested that the ``validate`` option always be set
        to ``True`` (the default) unless there is a specific need to read in
        invalid configuration from disk.

        Args:
           xmlData (String data): XML data representing configuration
           xmlPath (Absolute path to a file on disk): Path to an XML file on disk
           validate (Boolean true/false): Validate the document after parsing it
        Raises:
           ValueError: If both ``xmlData`` and ``xmlPath`` are passed-in
           ValueError: If the XML data in ``xmlData`` or ``xmlPath`` cannot be parsed
           ValueError: If the parsed configuration document is not valid
        """
        self._capacity = None
        self.capacity = None
        if xmlData is not None and xmlPath is not None:
            raise ValueError("Use either xmlData or xmlPath, but not both.")
        if xmlData is not None:
            self._parseXmlData(xmlData)
            if validate:
                self.validate()
        elif xmlPath is not None:
            with open(xmlPath) as f:  # pylint: disable=unspecified-encoding
                xmlData = f.read()
            self._parseXmlData(xmlData)
            if validate:
                self.validate()

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "LocalConfig(%s)" % (self.capacity)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Lists within this class are "unordered" for equality comparisons.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.capacity != other.capacity:
            if self.capacity < other.capacity:
                return -1
            else:
                return 1
        return 0

    def _setCapacity(self, value):
        """
        Property target used to set the capacity configuration value.
        If not ``None``, the value must be a ``CapacityConfig`` object.
        Raises:
           ValueError: If the value is not a ``CapacityConfig``
        """
        if value is None:
            self._capacity = None
        else:
            if not isinstance(value, CapacityConfig):
                raise ValueError("Value must be a ``CapacityConfig`` object.")
            self._capacity = value

    def _getCapacity(self):
        """
        Property target used to get the capacity configuration value.
        """
        return self._capacity

    capacity = property(_getCapacity, _setCapacity, None, "Capacity configuration in terms of a ``CapacityConfig`` object.")

    def validate(self):
        """
        Validates configuration represented by the object.
        THere must be either a percentage, or a byte capacity, but not both.
        Raises:
           ValueError: If one of the validations fails
        """
        if self.capacity is None:
            raise ValueError("Capacity section is required.")
        if self.capacity.maxPercentage is None and self.capacity.minBytes is None:
            raise ValueError("Must provide either max percentage or min bytes.")
        if self.capacity.maxPercentage is not None and self.capacity.minBytes is not None:
            raise ValueError("Must provide either max percentage or min bytes, but not both.")

    def addConfig(self, xmlDom, parentNode):
        """
        Adds a <capacity> configuration section as the next child of a parent.

        Third parties should use this function to write configuration related to
        this extension.

        We add the following fields to the document::

           maxPercentage  //cb_config/capacity/max_percentage
           minBytes       //cb_config/capacity/min_bytes

        Args:
           xmlDom: DOM tree as from ``impl.createDocument()``
           parentNode: Parent that the section should be appended to
        """
        if self.capacity is not None:
            sectionNode = addContainerNode(xmlDom, parentNode, "capacity")
            LocalConfig._addPercentageQuantity(xmlDom, sectionNode, "max_percentage", self.capacity.maxPercentage)
            if self.capacity.minBytes is not None:  # because utility function fills in empty section on None
                addByteQuantityNode(xmlDom, sectionNode, "min_bytes", self.capacity.minBytes)

    def _parseXmlData(self, xmlData):
        """
        Internal method to parse an XML string into the object.

        This method parses the XML document into a DOM tree (``xmlDom``) and then
        calls a static method to parse the capacity configuration section.

        Args:
           xmlData (String data): XML data to be parsed
        Raises:
           ValueError: If the XML cannot be successfully parsed
        """
        (xmlDom, parentNode) = createInputDom(xmlData)
        self._capacity = LocalConfig._parseCapacity(parentNode)

    @staticmethod
    def _parseCapacity(parentNode):
        """
        Parses a capacity configuration section.

        We read the following fields::

           maxPercentage  //cb_config/capacity/max_percentage
           minBytes       //cb_config/capacity/min_bytes

        Args:
           parentNode: Parent node to search beneath

        Returns:
            ``CapacityConfig`` object or ``None`` if the section does not exist
        Raises:
           ValueError: If some filled-in value is invalid
        """
        capacity = None
        section = readFirstChild(parentNode, "capacity")
        if section is not None:
            capacity = CapacityConfig()
            capacity.maxPercentage = LocalConfig._readPercentageQuantity(section, "max_percentage")
            capacity.minBytes = readByteQuantity(section, "min_bytes")
        return capacity

    @staticmethod
    def _readPercentageQuantity(parent, name):
        """
        Read a percentage quantity value from an XML document.
        Args:
           parent: Parent node to search beneath
           name: Name of node to search for
        Returns:
            Percentage quantity parsed from XML document
        """
        quantity = readString(parent, name)
        if quantity is None:
            return None
        return PercentageQuantity(quantity)

    @staticmethod
    def _addPercentageQuantity(xmlDom, parentNode, nodeName, percentageQuantity):
        """
        Adds a text node as the next child of a parent, to contain a percentage quantity.

        If the ``percentageQuantity`` is None, then no node will be created.

        Args:
           xmlDom: DOM tree as from ``impl.createDocument()``
           parentNode: Parent node to create child for
           nodeName: Name of the new container node
           percentageQuantity: PercentageQuantity object to put into the XML document

        Returns:
            Reference to the newly-created node
        """
        if percentageQuantity is not None:
            addStringNode(xmlDom, parentNode, nodeName, percentageQuantity.quantity)


########################################################################
# Public functions
########################################################################

###########################
# executeAction() function
###########################


# pylint: disable=W0613
def executeAction(configPath, options, config):
    """
    Executes the capacity action.

    Args:
       configPath (String representing a path on disk): Path to configuration file on disk
       options (Options object): Program command-line options
       config (Config object): Program configuration
    Raises:
       ValueError: Under many generic error conditions
       IOError: If there are I/O problems reading or writing files
    """
    logger.debug("Executing capacity extended action.")
    if config.options is None or config.store is None:
        raise ValueError("Cedar Backup configuration is not properly filled in.")
    local = LocalConfig(xmlPath=configPath)
    if config.store.checkMedia:
        checkMediaState(config.store)  # raises exception if media is not initialized
    capacity = createWriter(config).retrieveCapacity()
    logger.debug("Media capacity: %s", capacity)
    if local.capacity.maxPercentage is not None:
        if capacity.utilized > local.capacity.maxPercentage.percentage:
            logger.error(
                "Media has reached capacity limit of %s%%: %.2f%% utilized",
                local.capacity.maxPercentage.quantity,
                capacity.utilized,
            )
    else:
        if capacity.bytesAvailable < local.capacity.minBytes:
            logger.error(
                "Media has reached capacity limit of %s: only %s available",
                local.capacity.minBytes,
                displayBytes(capacity.bytesAvailable),
            )
    logger.info("Executed the capacity extended action successfully.")



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2007-2008,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Spans staged data among multiple discs
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Notes
########################################################################

"""
Spans staged data among multiple discs

This is the Cedar Backup span tool.  It is intended for use by people who stage
more data than can fit on a single disc.  It allows a user to split staged data
among more than one disc.  It can't be an extension because it requires user
input when switching media.

Most configuration is taken from the Cedar Backup configuration file,
specifically the store section.  A few pieces of configuration are taken
directly from the user.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules and constants
########################################################################

import logging
import os
import sys
import tempfile

from CedarBackup3.actions.constants import STORE_INDICATOR
from CedarBackup3.actions.store import writeIndicatorFile
from CedarBackup3.actions.util import createWriter, findDailyDirs
from CedarBackup3.cli import (
    DEFAULT_CONFIG,
    DEFAULT_LOGFILE,
    DEFAULT_MODE,
    DEFAULT_OWNERSHIP,
    Options,
    setupLogging,
    setupPathResolver,
)
from CedarBackup3.config import Config
from CedarBackup3.filesystem import BackupFileList, compareDigestMaps, normalizeDir
from CedarBackup3.release import AUTHOR, EMAIL, VERSION
from CedarBackup3.util import UNIT_BYTES, UNIT_SECTORS, Diagnostics, convertSize, displayBytes, mount, unmount

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.tools.span")


#######################################################################
# SpanOptions class
#######################################################################


class SpanOptions(Options):
    """
    Tool-specific command-line options.

    Most of the cback3 command-line options are exactly what we need here --
    logfile path, permissions, verbosity, etc.  However, we need to make a few
    tweaks since we don't accept any actions.

    Also, a few extra command line options that we accept are really ignored
    underneath.  I just don't care about that for a tool like this.
    """

    def validate(self):
        """
        Validates command-line options represented by the object.
        There are no validations here, because we don't use any actions.
        Raises:
           ValueError: If one of the validations fails
        """
        pass


#######################################################################
# Public functions
#######################################################################

#################
# cli() function
#################


def cli():
    """
    Implements the command-line interface for the ``cback3-span`` script.

    Essentially, this is the "main routine" for the cback3-span script.  It does
    all of the argument processing for the script, and then also implements the
    tool functionality.

    This function looks pretty similiar to ``CedarBackup3.cli.cli()``.  It's not
    easy to refactor this code to make it reusable and also readable, so I've
    decided to just live with the duplication.

    A different error code is returned for each type of failure:

       - ``1``: The Python interpreter version is not supported
       - ``2``: Error processing command-line arguments
       - ``3``: Error configuring logging
       - ``4``: Error parsing indicated configuration file
       - ``5``: Backup was interrupted with a CTRL-C or similar
       - ``6``: Error executing other parts of the script

    *Note:* This script uses print rather than logging to the INFO level, because
    it is interactive.  Underlying Cedar Backup functionality uses the logging
    mechanism exclusively.

    Returns:
        Error code as described above
    """
    try:
        if list(map(int, [sys.version_info[0], sys.version_info[1]])) < [3, 8]:
            sys.stderr.write("Python 3 version 3.8 or greater required.\n")
            return 1
    except:
        # sys.version_info isn't available before 2.0
        sys.stderr.write("Python 3 version 3.8 or greater required.\n")
        return 1

    try:
        options = SpanOptions(argumentList=sys.argv[1:])
    except Exception as e:
        _usage()
        sys.stderr.write(" *** Error: %s\n" % e)
        return 2

    if options.help:
        _usage()
        return 0
    if options.version:
        _version()
        return 0
    if options.diagnostics:
        _diagnostics()
        return 0

    if options.stacktrace:
        logfile = setupLogging(options)
    else:
        try:
            logfile = setupLogging(options)
        except Exception as e:
            sys.stderr.write("Error setting up logging: %s\n" % e)
            return 3

    logger.info("Cedar Backup 'span' utility run started.")
    logger.info("Options were [%s]", options)
    logger.info("Logfile is [%s]", logfile)

    if options.config is None:
        logger.debug("Using default configuration file.")
        configPath = DEFAULT_CONFIG
    else:
        logger.debug("Using user-supplied configuration file.")
        configPath = options.config

    try:
        logger.info("Configuration path is [%s]", configPath)
        config = Config(xmlPath=configPath)
        setupPathResolver(config)
    except Exception as e:
        logger.error("Error reading or handling configuration: %s", e)
        logger.info("Cedar Backup 'span' utility run completed with status 4.")
        return 4

    if options.stacktrace:
        _executeAction(options, config)
    else:
        try:
            _executeAction(options, config)
        except KeyboardInterrupt:
            logger.error("Backup interrupted.")
            logger.info("Cedar Backup 'span' utility run completed with status 5.")
            return 5
        except Exception as e:
            logger.error("Error executing backup: %s", e)
            logger.info("Cedar Backup 'span' utility run completed with status 6.")
            return 6

    logger.info("Cedar Backup 'span' utility run completed with status 0.")
    return 0


#######################################################################
# Utility functions
#######################################################################

####################
# _usage() function
####################


def _usage(fd=sys.stderr):
    """
    Prints usage information for the cback3-span script.
    Args:
       fd: File descriptor used to print information
    *Note:* The ``fd`` is used rather than ``print`` to facilitate unit testing.
    """
    fd.write("\n")
    fd.write(" Usage: cback3-span [switches]\n")
    fd.write("\n")
    fd.write(" Cedar Backup 'span' tool.\n")
    fd.write("\n")
    fd.write(" This Cedar Backup utility spans staged data between multiple discs.\n")
    fd.write(" It is a utility, not an extension, and requires user interaction.\n")
    fd.write("\n")
    fd.write(" The following switches are accepted, mostly to set up underlying\n")
    fd.write(" Cedar Backup functionality:\n")
    fd.write("\n")
    fd.write("   -h, --help     Display this usage/help listing\n")
    fd.write("   -V, --version  Display version information\n")
    fd.write("   -b, --verbose  Print verbose output as well as logging to disk\n")
    fd.write("   -c, --config   Path to config file (default: %s)\n" % DEFAULT_CONFIG)
    fd.write("   -l, --logfile  Path to logfile (default: %s)\n" % DEFAULT_LOGFILE)
    fd.write("   -o, --owner    Logfile ownership, user:group (default: %s:%s)\n" % (DEFAULT_OWNERSHIP[0], DEFAULT_OWNERSHIP[1]))
    fd.write("   -m, --mode     Octal logfile permissions mode (default: %o)\n" % DEFAULT_MODE)
    fd.write("   -O, --output   Record some sub-command (i.e. tar) output to the log\n")
    fd.write("   -d, --debug    Write debugging information to the log (implies --output)\n")
    fd.write("   -s, --stack    Dump a Python stack trace instead of swallowing exceptions\n")
    fd.write("\n")


######################
# _version() function
######################


def _version(fd=sys.stdout):
    """
    Prints version information for the cback3-span script.
    Args:
       fd: File descriptor used to print information
    *Note:* The ``fd`` is used rather than ``print`` to facilitate unit testing.
    """
    fd.write("\n")
    fd.write(" Cedar Backup 'span' tool.\n")
    fd.write(" Included with Cedar Backup version %s.\n" % VERSION)
    fd.write("\n")
    fd.write(" Copyright (c) %s <%s>.\n" % (AUTHOR, EMAIL))
    fd.write(" See NOTICE for a list of included code and other contributors.\n")
    fd.write(" This is free software; there is NO warranty.  See the\n")
    fd.write(" GNU General Public License version 2 for copying conditions.\n")
    fd.write("\n")
    fd.write(" Use the --help option for usage information.\n")
    fd.write("\n")


##########################
# _diagnostics() function
##########################


def _diagnostics(fd=sys.stdout):
    """
    Prints runtime diagnostics information.
    Args:
       fd: File descriptor used to print information
    *Note:* The ``fd`` is used rather than ``print`` to facilitate unit testing.
    """
    fd.write("\n")
    fd.write("Diagnostics:\n")
    fd.write("\n")
    Diagnostics().printDiagnostics(fd=fd, prefix="   ")
    fd.write("\n")


############################
# _executeAction() function
############################


# pylint: disable=W0613
def _executeAction(options, config):
    """
    Implements the guts of the cback3-span tool.

    Args:
       options (SpanOptions object): Program command-line options
       config (Config object): Program configuration
    Raises:
       Exception: Under many generic error conditions
    """
    print("")
    print("================================================")
    print("           Cedar Backup 'span' tool")
    print("================================================")
    print("")
    print("This the Cedar Backup span tool.  It is used to split up staging")
    print("data when that staging data does not fit onto a single disc.")
    print("")
    print("This utility operates using Cedar Backup configuration.  Configuration")
    print("specifies which staging directory to look at and which writer device")
    print("and media type to use.")
    print("")
    if not _getYesNoAnswer("Continue?", default="Y"):
        return
    print("===")

    print("")
    print("Cedar Backup store configuration looks like this:")
    print("")
    print("   Source Directory...: %s" % config.store.sourceDir)
    print("   Media Type.........: %s" % config.store.mediaType)
    print("   Device Type........: %s" % config.store.deviceType)
    print("   Device Path........: %s" % config.store.devicePath)
    print("   Device SCSI ID.....: %s" % config.store.deviceScsiId)
    print("   Drive Speed........: %s" % config.store.driveSpeed)
    print("   Check Data Flag....: %s" % config.store.checkData)
    print("   No Eject Flag......: %s" % config.store.noEject)
    print("")
    if not _getYesNoAnswer("Is this OK?", default="Y"):
        return
    print("===")

    (writer, mediaCapacity) = _getWriter(config)

    print("")
    print("Please wait, indexing the source directory (this may take a while)...")
    (dailyDirs, fileList) = _findDailyDirs(config.store.sourceDir)
    print("===")

    print("")
    print("The following daily staging directories have not yet been written to disc:")
    print("")
    for dailyDir in dailyDirs:
        print("   %s" % dailyDir)

    totalSize = fileList.totalSize()
    print("")
    print("The total size of the data in these directories is %s." % displayBytes(totalSize))
    print("")
    if not _getYesNoAnswer("Continue?", default="Y"):
        return
    print("===")

    print("")
    print("Based on configuration, the capacity of your media is %s." % displayBytes(mediaCapacity))

    print("")
    print("Since estimates are not perfect and there is some uncertainly in")
    print('media capacity calculations, it is good to have a "cushion",')
    print("a percentage of capacity to set aside.  The cushion reduces the")
    print("capacity of your media, so a 1.5% cushion leaves 98.5% remaining.")
    print("")
    cushion = _getFloat("What cushion percentage?", default=4.5)
    print("===")

    realCapacity = ((100.0 - cushion) / 100.0) * mediaCapacity
    minimumDiscs = (totalSize / realCapacity) + 1
    print("")
    print("The real capacity, taking into account the %.2f%% cushion, is %s." % (cushion, displayBytes(realCapacity)))
    print("It will take at least %d disc(s) to store your %s of data." % (minimumDiscs, displayBytes(totalSize)))
    print("")
    if not _getYesNoAnswer("Continue?", default="Y"):
        return
    print("===")

    happy = False
    while not happy:
        print("")
        print("Which algorithm do you want to use to span your data across")
        print("multiple discs?")
        print("")
        print("The following algorithms are available:")
        print("")
        print('   first....: The "first-fit" algorithm')
        print('   best.....: The "best-fit" algorithm')
        print('   worst....: The "worst-fit" algorithm')
        print('   alternate: The "alternate-fit" algorithm')
        print("")
        print("If you don't like the results you will have a chance to try a")
        print("different one later.")
        print("")
        algorithm = _getChoiceAnswer("Which algorithm?", "worst", ["first", "best", "worst", "alternate"])
        print("===")

        print("")
        print("Please wait, generating file lists (this may take a while)...")
        spanSet = fileList.generateSpan(capacity=realCapacity, algorithm="%s_fit" % algorithm)
        print("===")

        print("")
        print('Using the "%s-fit" algorithm, Cedar Backup can split your data' % algorithm)
        print("into %d discs." % len(spanSet))
        print("")
        counter = 0
        for item in spanSet:
            counter += 1
            print(
                "Disc %d: %d files, %s, %.2f%% utilization"
                % (counter, len(item.fileList), displayBytes(item.size), item.utilization)
            )
        print("")
        if _getYesNoAnswer("Accept this solution?", default="Y"):
            happy = True
        print("===")

    counter = 0
    for spanItem in spanSet:
        counter += 1
        if counter == 1:
            print("")
            _getReturn("Please place the first disc in your backup device.\nPress return when ready.")
            print("===")
        else:
            print("")
            _getReturn("Please replace the disc in your backup device.\nPress return when ready.")
            print("===")
        _writeDisc(config, writer, spanItem)

    _writeStoreIndicator(config, dailyDirs)

    print("")
    print("Completed writing all discs.")


############################
# _findDailyDirs() function
############################


def _findDailyDirs(stagingDir):
    """
    Returns a list of all daily staging directories that have not yet been
    stored.

    The store indicator file ``cback.store`` will be written to a daily staging
    directory once that directory is written to disc.  So, this function looks
    at each daily staging directory within the configured staging directory, and
    returns a list of those which do not contain the indicator file.

    Returned is a tuple containing two items: a list of daily staging
    directories, and a BackupFileList containing all files among those staging
    directories.

    Args:
       stagingDir: Configured staging directory

    Returns:
        Tuple (staging dirs, backup file list)
    """
    results = findDailyDirs(stagingDir, STORE_INDICATOR)
    fileList = BackupFileList()
    for item in results:
        fileList.addDirContents(item)
    return (results, fileList)


##################################
# _writeStoreIndicator() function
##################################


def _writeStoreIndicator(config, dailyDirs):
    """
    Writes a store indicator file into daily directories.

    Args:
       config: Config object
       dailyDirs: List of daily directories
    """
    for dailyDir in dailyDirs:
        writeIndicatorFile(dailyDir, STORE_INDICATOR, config.options.backupUser, config.options.backupGroup)


########################
# _getWriter() function
########################


def _getWriter(config):
    """
    Gets a writer and media capacity from store configuration.
    Returned is a writer and a media capacity in bytes.
    Args:
       config: Cedar Backup configuration
    Returns:
        Tuple of (writer, mediaCapacity)
    """
    writer = createWriter(config)
    mediaCapacity = convertSize(writer.media.capacity, UNIT_SECTORS, UNIT_BYTES)
    return (writer, mediaCapacity)


########################
# _writeDisc() function
########################


def _writeDisc(config, writer, spanItem):
    """
    Writes a span item to disc.
    Args:
       config: Cedar Backup configuration
       writer: Writer to use
       spanItem: Span item to write
    """
    print("")
    _discInitializeImage(config, writer, spanItem)
    _discWriteImage(config, writer)
    _discConsistencyCheck(config, writer, spanItem)
    print("Write process is complete.")
    print("===")


def _discInitializeImage(config, writer, spanItem):
    """
    Initialize an ISO image for a span item.
    Args:
       config: Cedar Backup configuration
       writer: Writer to use
       spanItem: Span item to write
    """
    complete = False
    while not complete:
        try:
            print("Initializing image...")
            writer.initializeImage(newDisc=True, tmpdir=config.options.workingDir)
            for path in spanItem.fileList:
                graftPoint = os.path.dirname(path.replace(config.store.sourceDir, "", 1))
                writer.addImageEntry(path, graftPoint)
            complete = True
        except KeyboardInterrupt as e:
            raise e
        except Exception as e:
            logger.error("Failed to initialize image: %s", e)
            if not _getYesNoAnswer("Retry initialization step?", default="Y"):
                raise e
            print("Ok, attempting retry.")
            print("===")
    print("Completed initializing image.")


# pylint: disable=W0613
def _discWriteImage(config, writer):
    """
    Writes a ISO image for a span item.
    Args:
       config: Cedar Backup configuration
       writer: Writer to use
    """
    complete = False
    while not complete:
        try:
            print("Writing image to disc...")
            writer.writeImage()
            complete = True
        except KeyboardInterrupt as e:
            raise e
        except Exception as e:
            logger.error("Failed to write image: %s", e)
            if not _getYesNoAnswer("Retry this step?", default="Y"):
                raise e
            print("Ok, attempting retry.")
            _getReturn("Please replace media if needed.\nPress return when ready.")
            print("===")
    print("Completed writing image.")


def _discConsistencyCheck(config, writer, spanItem):
    """
    Run a consistency check on an ISO image for a span item.
    Args:
       config: Cedar Backup configuration
       writer: Writer to use
       spanItem: Span item to write
    """
    if config.store.checkData:
        complete = False
        while not complete:
            try:
                print("Running consistency check...")
                _consistencyCheck(config, spanItem.fileList)
                complete = True
            except KeyboardInterrupt as e:
                raise e
            except Exception as e:
                logger.error("Consistency check failed: %s", e)
                if not _getYesNoAnswer("Retry the consistency check?", default="Y"):
                    raise e
                if _getYesNoAnswer("Rewrite the disc first?", default="N"):
                    print("Ok, attempting retry.")
                    _getReturn("Please replace the disc in your backup device.\nPress return when ready.")
                    print("===")
                    _discWriteImage(config, writer)
                else:
                    print("Ok, attempting retry.")
                    print("===")
        print("Completed consistency check.")


###############################
# _consistencyCheck() function
###############################


def _consistencyCheck(config, fileList):
    """
    Runs a consistency check against media in the backup device.

    The function mounts the device at a temporary mount point in the working
    directory, and then compares the passed-in file list's digest map with the
    one generated from the disc.  The two lists should be identical.

    If no exceptions are thrown, there were no problems with the consistency
    check.

    @warning: The implementation of this function is very UNIX-specific.

    Args:
       config: Config object
       fileList: BackupFileList whose contents to check against

    Raises:
       ValueError: If the check fails
       IOError: If there is a problem working with the media
    """
    logger.debug("Running consistency check.")
    mountPoint = tempfile.mkdtemp(dir=config.options.workingDir)
    try:
        mount(config.store.devicePath, mountPoint, "iso9660")
        discList = BackupFileList()
        discList.addDirContents(mountPoint)
        sourceList = BackupFileList()
        sourceList.extend(fileList)
        discListDigest = discList.generateDigestMap(stripPrefix=normalizeDir(mountPoint))
        sourceListDigest = sourceList.generateDigestMap(stripPrefix=normalizeDir(config.store.sourceDir))
        compareDigestMaps(sourceListDigest, discListDigest, verbose=True)
        logger.info("Consistency check completed.  No problems found.")
    finally:
        unmount(mountPoint, True, 5, 1)  # try 5 times, and remove mount point when done


#########################################################################
# User interface utilities
########################################################################


def _getYesNoAnswer(prompt, default):
    """
    Get a yes/no answer from the user.
    The default will be placed at the end of the prompt.
    A "Y" or "y" is considered yes, anything else no.
    A blank (empty) response results in the default.
    Args:
       prompt: Prompt to show
       default: Default to set if the result is blank
    Returns:
        Boolean true/false corresponding to Y/N
    """
    if default == "Y":
        prompt = "%s [Y/n]: " % prompt
    else:
        prompt = "%s [y/N]: " % prompt
    answer = input(prompt)
    if answer in [None, ""]:
        answer = default
    if answer[0] in ["Y", "y"]:
        return True
    else:
        return False


def _getChoiceAnswer(prompt, default, validChoices):
    """
    Get a particular choice from the user.
    The default will be placed at the end of the prompt.
    The function loops until getting a valid choice.
    A blank (empty) response results in the default.
    Args:
       prompt: Prompt to show
       default: Default to set if the result is None or blank
       validChoices: List of valid choices (strings)
    Returns:
        Valid choice from user
    """
    prompt = "%s [%s]: " % (prompt, default)
    answer = input(prompt)
    if answer in [None, ""]:
        answer = default
    while answer not in validChoices:
        print("Choice must be one of %s" % validChoices)
        answer = input(prompt)
    return answer


def _getFloat(prompt, default):
    """
    Get a floating point number from the user.
    The default will be placed at the end of the prompt.
    The function loops until getting a valid floating point number.
    A blank (empty) response results in the default.
    Args:
       prompt: Prompt to show
       default: Default to set if the result is None or blank
    Returns:
        Floating point number from user
    """
    prompt = "%s [%.2f]: " % (prompt, default)
    while True:
        answer = input(prompt)
        if answer in [None, ""]:
            return default
        else:
            try:
                return float(answer)
            except ValueError:
                print("Enter a floating point number.")


def _getReturn(prompt):
    """
    Get a return key from the user.
    Args:
       prompt: Prompt to show
    """
    input(prompt)


#########################################################################
# Main routine
########################################################################

if __name__ == "__main__":
    sys.exit(cli())
